static void 
#ifdef _CLOVER
assign_mult_eoprec_wd(const float        *kappa2,
                            cuhqfield       *yde,
                      const cuhqfield        *ye,
                      const cuhgvfield       *ue,
                      const cuhgvfield       *uo,
                      const int             *ieo,
                      const cuhclvfield *fclinve,
                      const cuhclvfield *fclinvo)
#else
assign_mult_eoprec_wd(const float      *kappa2,
                            cuhqfield     *yde,
                      const cuhqfield      *ye,
                      const cuhgvfield     *ue,
                      const cuhgvfield     *uo,
                      const int           *ieo)
#endif
{
///////////////////////////////////////////////////////////////////////////////
// 
// Multiply even/odd-site preconditioned Dirac operator : 
//       v
// yde = Dee ye,
//       v
// where Dee = 1 - kappa^2 Fee Meo Foo Moe
//
//  kappa2 : hopping parameter squared (kappa^2)
//      ye : even/odd site fermion (input)
//     yde : even/odd site fermion (output)
//      ue : even/odd site gauge link
//      uo : odd/even site gauge link
//     ieo : 0 or 1 (0:even,1:odd)
// fclinve : even/odd site inverse clover term
// fclinvo : odd/even site inverse clover term
///////////////////////////////////////////////////////////////////////////////
  cuhqfield *Myo;
  int *ioe;

  CUDA_SAFE_CALL(cudaMalloc((void**)&Myo,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioe,sizeof(float)));
  cureverse_eo(ieo,ioe); // ioe = (ieo+1) % 2

#ifdef _CLOVER
  // Myo = F^-1_oo Moe ye
  // yde = F^-1_ee Meo Myo = F^-1_ee Meo F^-1_oo Moe ye
  assign_mult_hop(Myo, ye,uo,ue,ioe,fclinvo);
  assign_mult_hop(yde,Myo,ue,uo,ieo,fclinve);
#else
  // Myo = Moe ye
  // yde = Meo Myo = Meo Moe ye
  assign_mult_hop(Myo, ye,uo,ue,ioe);
  assign_mult_hop(yde,Myo,ue,uo,ieo);
#endif

  //
  // yde = - kappa^2 yde + ye
  //
  accum_nmult_add(yde,kappa2,ye);

  CUDA_SAFE_CALL(cudaFree(Myo));
  CUDA_SAFE_CALL(cudaFree(ioe));
}

static void 
#ifdef _CLOVER
assign_mult_eoprec_dag_wd(const float        *kappa2,
                                cuhqfield       *yde,
                          const cuhqfield        *ye,
                          const cuhgvfield       *ue,
                          const cuhgvfield       *uo,
                          const int             *ieo,
                          const cuhclvfield *fclinve,
                          const cuhclvfield *fclinvo)
#else
assign_mult_eoprec_dag_wd(const float      *kappa2,
                                cuhqfield     *yde,
                          const cuhqfield      *ye,
                          const cuhgvfield     *ue,
                          const cuhgvfield     *uo,
                          const int           *ieo)
#endif
{
///////////////////////////////////////////////////////////////////////////////
// 
// Multiply dagger of even/odd-site preconditioned Dirac operator : 
//       v
// yde = Dee^+ ye,
//       v
// where Dee^+ = 1 - kappa^2 gamma5 Meo Foo Moe Fee gamma5
//
//  kappa2 : hopping parameter squared (kappa^2)
//      ye : even/odd site fermion (input)
//     yde : even/odd site fermion (output)
//      ue : even/odd site gauge link
//      uo : odd/even site gauge link
//     ieo : 0 or 1 (0:even,1:odd)
// fclinve : even/odd site inverse clover term
// fclinvo : odd/even site inverse clover term
///////////////////////////////////////////////////////////////////////////////
  cuhqfield *Myo;
  int *ioe;

  CUDA_SAFE_CALL(cudaMalloc((void**)&Myo,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioe,sizeof(float)));
  cureverse_eo(ieo,ioe); // ioe = (ieo+1) % 2

#ifdef _CLOVER
  // yde = gamma5 Fee ye
  assign_mult_clover_term(yde,ye,ieo,fclinve);
  accum_mult_gamma5(yde);

  // Myo = Foo Moe yde
  // yde = gamma5 Meo Myo = gamma5 Meo Foo Moe gamma5 Fee ye
  assign_mult_hop(Myo,yde,uo,ue,ioe,fclinvo);
  assign_mult_hop(yde,Myo,ue,uo,ieo);
  accum_mult_gamma5(yde);
#else
  // yde = gamma5 ye
  assign_mult_gamma5(yde,ye);

  // Myo = Moe yde
  // yde = gamma5 Meo Myo = gamma5 Meo Moe gamma5 ye
  assign_mult_hop(Myo,yde,uo,ue,ioe);
  assign_mult_hop(yde,Myo,ue,uo,ieo);
  accum_mult_gamma5(yde);
#endif

  //
  // yde = - kappa^2 yde + ye
  //
  accum_nmult_add(yde,kappa2,ye);

  CUDA_SAFE_CALL(cudaFree(Myo));
  CUDA_SAFE_CALL(cudaFree(ioe));
}
