/*
*  $Id: qmx_lib.c,v 1.9 2011/08/16 13:37:36 ishikawa Exp $
*  QMX: library file for QCD communication API on top of the Open-MX API.
*/

#include "qmx_lib.h"
#undef _CHECKSUM_
static long long int qmx_call_counter = 0;

static const int QMX_NUM_RANK = 4;
static const int QMX_NUM_CH = 2;
static const int QMX_HNAME_LEN = 256;
static const int QMX_OPEN_WAITTIME = 1; // in sec //
static const int QMX_RING_ASCEND  = +1;
static const int QMX_RING_DESCEND = -1;
static const int QMX_CHANNEL_ERROR = -1;

static const uint32_t QMX_FILTER  = 0xDEADBEEF;
static const uint32_t QMX_TIMEOUT = MX_INFINITE;

typedef struct __qmx_channel_t {
  int  local_rank;
  int remote_rank;
  uint64_t  local_nic_id;
  uint64_t remote_nic_id;
  uint32_t   local_ep_id;
  uint32_t  remote_ep_id;
  mx_endpoint_t       local_ep;
  mx_endpoint_addr_t  remote_ep_addr;
  uint32_t filter;
} qmx_channel_t __attribute__ ((aligned (16)));

static unsigned int  qmx_local_rank;
static char          qmx_hname[QMX_HNAME_LEN];
static qmx_channel_t qmx_ch[QMX_NUM_CH];

static void qmx_setup_table();
static void qmx_connect();
static void qmx_ring_transmission(const int direction, const int byte, void *sbuff, void *rbuff);

static inline unsigned int 
__qmx_get_local_rank() { return qmx_local_rank; }

static inline unsigned int 
__qmx_get_rank_x(const unsigned int rank) {
  unsigned int rank_x = (2 & rank) >> 1;
  return rank_x;
}

static inline unsigned int 
__qmx_get_rank_y(const unsigned int rank) {
  unsigned int rank_y = (1 & rank);
  return rank_y;
}

static inline unsigned int 
__qmx_get_parity(const unsigned int rank) {
  unsigned int rank_x = __qmx_get_rank_x(rank);
  unsigned int rank_y = __qmx_get_rank_y(rank);
  unsigned int parity = rank_x ^ rank_y;
  return parity;
}

static inline unsigned int 
__qmx_get_prev_rank(const unsigned int rank) {
  unsigned int parity = __qmx_get_parity(rank);
  return rank ^ (1 << parity);
}

static inline unsigned int 
__qmx_get_next_rank(const unsigned int rank) {
  unsigned int parity = __qmx_get_parity(rank);
  return rank ^ (1 << (1 ^ parity));
}

static inline unsigned int 
__qmx_get_1d_next_rank(const unsigned int rank){
  return __qmx_get_next_rank(rank);
}

static inline unsigned int 
__qmx_get_1d_prev_rank(const unsigned int rank){
  return __qmx_get_prev_rank(rank);
}

static inline unsigned int 
__qmx_get_2d_next_x_rank(const unsigned int rank){
  unsigned int rank_x = __qmx_get_rank_x(rank);
  unsigned int rank_y = __qmx_get_rank_y(rank);
  rank_x = (rank_x ^ 1);
  return (rank_x << 1) ^ rank_y;
}

static inline unsigned int 
__qmx_get_2d_next_y_rank(const unsigned int rank){
  unsigned int rank_x = __qmx_get_rank_x(rank);
  unsigned int rank_y = __qmx_get_rank_y(rank);
  rank_y = (rank_y ^ 1);
  return (rank_x << 1) ^ rank_y;
}

static inline unsigned int 
__qmx_get_2d_prev_x_rank(const unsigned int rank){
  return __qmx_get_2d_next_x_rank(rank);
}

static inline unsigned int 
__qmx_get_2d_prev_y_rank(const unsigned int rank){
  return __qmx_get_2d_next_y_rank(rank);
}

static inline long long int
__qmx_get_checksum(const int length, char *vec) {
  long long int csum;
  csum = 0;
  for (int i=0; i < length; ++i) {
    csum = csum + (long long int)(vec[i]);
  }
  return csum;
}

static inline void 
rc_check(mx_return_t *rc, const char *name){
#ifdef _DEBUG_
  fprintf(stderr,"%s : %s : %s\n",qmx_hname,name,mx_strerror(*rc));
#endif
  if (MX_SUCCESS != *rc) {
    fprintf(stderr,"FAIL: %s : %s : %s\n",qmx_hname,name,mx_strerror(*rc));
    mx_finalize();
    exit(EXIT_FAILURE);
  } 
} 

extern "C"
int qmx_get_rank_size() { return QMX_NUM_RANK; }

extern "C"
int qmx_get_local_rank() { return qmx_local_rank; }

extern "C"
int qmx_get_2d_local_rank_x() { return __qmx_get_rank_x(qmx_local_rank); }

extern "C"
int qmx_get_2d_local_rank_y() { return __qmx_get_rank_y(qmx_local_rank); }

extern "C"
int qmx_get_1d_next_rank() { return __qmx_get_next_rank(qmx_local_rank); }

extern "C"
int qmx_get_1d_prev_rank() { return __qmx_get_prev_rank(qmx_local_rank); }

extern "C"
int qmx_get_2d_next_x_rank() { return __qmx_get_2d_next_x_rank(qmx_local_rank); }

extern "C"
int qmx_get_2d_next_y_rank() { return __qmx_get_2d_next_y_rank(qmx_local_rank); }

extern "C"
int qmx_get_2d_prev_x_rank() { return __qmx_get_2d_prev_x_rank(qmx_local_rank); }

extern "C"
int qmx_get_2d_prev_y_rank() { return __qmx_get_2d_prev_y_rank(qmx_local_rank); }


extern "C"
void qmx_clock(double *time) {
  struct timespec t;
  clock_gettime(CLOCK_REALTIME,&t);
  *time = ((double)(t.tv_sec*1000000000ULL+t.tv_nsec))/1000000000.0;
}

static inline int 
qmx_get_channel_id(int remote_rank) {
  for (int ich = 0; ich < QMX_NUM_CH; ich++) {
    if ( remote_rank == qmx_ch[ich].remote_rank ) {
      return ich;
    }
  }
  fprintf(stderr,"No channel to the destination rank %d from local rank %d.\n",remote_rank,qmx_local_rank);
  return QMX_CHANNEL_ERROR;
}

static void 
qmx_setup_table() {

  gethostname(qmx_hname,QMX_HNAME_LEN);

  qmx_ch[0].filter = QMX_FILTER+0;
  qmx_ch[1].filter = QMX_FILTER+1;

  if (0 == strcmp(qmx_hname,"gameqcd01")) {

    qmx_local_rank = 0;

    qmx_ch[0].local_rank    = 0;
    qmx_ch[0].local_nic_id  = 0x001b21944db6L;
    qmx_ch[0].local_ep_id   = 3;

    qmx_ch[1].local_rank    = 0;
    qmx_ch[1].local_nic_id  = 0x001b21944db7L;
    qmx_ch[1].local_ep_id   = 4;

    qmx_ch[0].remote_rank   = 2;
    qmx_ch[0].remote_nic_id = 0x001b21944d99L;
    qmx_ch[0].remote_ep_id  = 3;

    qmx_ch[1].remote_rank   = 1;
    qmx_ch[1].remote_nic_id = 0x001b21944e22L;
    qmx_ch[1].remote_ep_id  = 4;

  } else if (0 == strcmp(qmx_hname,"gameqcd02")) {

    qmx_local_rank = 1;

    qmx_ch[0].local_rank    = 1;
    qmx_ch[0].local_nic_id  = 0x001b21944e23L;
    qmx_ch[0].local_ep_id   = 3;

    qmx_ch[1].local_rank    = 1;
    qmx_ch[1].local_nic_id  = 0x001b21944e22L;
    qmx_ch[1].local_ep_id   = 4;

    qmx_ch[0].remote_rank   = 3;
    qmx_ch[0].remote_nic_id = 0x001b21944e18L;
    qmx_ch[0].remote_ep_id  = 3;

    qmx_ch[1].remote_rank   = 0;
    qmx_ch[1].remote_nic_id = 0x001b21944db7L;
    qmx_ch[1].remote_ep_id  = 4;

  } else if (0 == strcmp(qmx_hname,"gameqcd03")) {

    qmx_local_rank = 2;

    qmx_ch[0].local_rank    = 2;
    qmx_ch[0].local_nic_id  = 0x001b21944d99L;
    qmx_ch[0].local_ep_id   = 3;

    qmx_ch[1].local_rank    = 2;
    qmx_ch[1].local_nic_id  = 0x001b21944d98L;
    qmx_ch[1].local_ep_id   = 4;

    qmx_ch[0].remote_rank   = 0;
    qmx_ch[0].remote_nic_id = 0x001b21944db6L;
    qmx_ch[0].remote_ep_id  = 3;

    qmx_ch[1].remote_rank   = 3;
    qmx_ch[1].remote_nic_id = 0x001b21944e19L;
    qmx_ch[1].remote_ep_id  = 4;

  } else if (0 == strcmp(qmx_hname,"gameqcd04")) {

    qmx_local_rank = 3;

    qmx_ch[0].local_rank    = 3;
    qmx_ch[0].local_nic_id  = 0x001b21944e18L;
    qmx_ch[0].local_ep_id   = 3;

    qmx_ch[1].local_rank    = 3;
    qmx_ch[1].local_nic_id  = 0x001b21944e19L;
    qmx_ch[1].local_ep_id   = 4;

    qmx_ch[0].remote_rank   = 1;
    qmx_ch[0].remote_nic_id = 0x001b21944e23L;
    qmx_ch[0].remote_ep_id  = 3;

    qmx_ch[1].remote_rank   = 2;
    qmx_ch[1].remote_nic_id = 0x001b21944d98L;
    qmx_ch[1].remote_ep_id  = 4;

  } else {
    fprintf(stderr,"No matching host name. too many hosts?\n");
    exit(EXIT_FAILURE);
  }
}

static void 
qmx_connect() {
  mx_return_t rc;
  mx_segment_t recv_seg,send_seg;
  mx_request_t recv_req,send_req;
  mx_status_t status;
  uint32_t result;

  int sbuff;
  int rbuff;

  for (int ich = 0; ich < QMX_NUM_CH; ich++) {

    sbuff = qmx_ch[ich].local_rank;
    rbuff = 0;
    send_seg.segment_ptr = &sbuff;
    send_seg.segment_length = sizeof(sbuff);
    recv_seg.segment_ptr = &rbuff;
    recv_seg.segment_length = sizeof(rbuff);

    if ( qmx_ch[ich].remote_rank > qmx_ch[ich].local_rank ) {

      rc = mx_connect( qmx_ch[ich].local_ep,
                       qmx_ch[ich].remote_nic_id,
                       qmx_ch[ich].remote_ep_id,
                       qmx_ch[ich].filter, QMX_TIMEOUT,
                     &(qmx_ch[ich].remote_ep_addr));
      rc_check(&rc,"mx_connect");

      rc = mx_issend( qmx_ch[ich].local_ep, &send_seg, 1, qmx_ch[ich].remote_ep_addr, 0x0L, NULL, &send_req);
      rc_check(&rc,"mx_isend");
      rc =   mx_wait( qmx_ch[ich].local_ep, &send_req, QMX_TIMEOUT, &status, &result);
      rc_check(&rc,"mx_wait");

      rc = mx_irecv( qmx_ch[ich].local_ep, &recv_seg, 1, 0x0L, 0x0L, NULL, &recv_req);
      rc_check(&rc,"mx_irecv");
      rc =  mx_wait( qmx_ch[ich].local_ep, &recv_req, QMX_TIMEOUT, &status, &result);
      rc_check(&rc,"mx_wait");

    } else {

      rc = mx_irecv(qmx_ch[ich].local_ep,&recv_seg,1,0x0L,0x0L,NULL,&recv_req);
      rc_check(&rc,"mx_irecv");
      rc =  mx_wait(qmx_ch[ich].local_ep,&recv_req,QMX_TIMEOUT,&status,&result);
      rc_check(&rc,"mx_wait");

      rc = mx_connect( qmx_ch[ich].local_ep,
                       qmx_ch[ich].remote_nic_id,
                       qmx_ch[ich].remote_ep_id,
                       qmx_ch[ich].filter, QMX_TIMEOUT,
                     &(qmx_ch[ich].remote_ep_addr));
      rc_check(&rc,"mx_connect");

      rc = mx_issend( qmx_ch[ich].local_ep, &send_seg, 1, qmx_ch[ich].remote_ep_addr, 0x0L, NULL, &send_req);
      rc_check(&rc,"mx_isend");
      rc =   mx_wait( qmx_ch[ich].local_ep, &send_req, QMX_TIMEOUT, &status, &result);
      rc_check(&rc,"mx_wait");

    }
#ifdef _DEBUG_
    fprintf(stderr,"%d %d -> %d, X -> %d\n",qmx_ch[ich].local_rank , sbuff,
                                            qmx_ch[ich].remote_rank, rbuff);
#endif
  }
}

extern "C"
void qmx_init() {
  mx_return_t rc;

  rc = mx_init();
  rc_check(&rc,"mx_init");

  qmx_setup_table();

  for (int ich = 0; ich < QMX_NUM_CH; ich++) {
    int board = ich + 3;
    rc = mx_open_endpoint( board, qmx_ch[ich].local_ep_id,
                                  qmx_ch[ich].filter, NULL,  0,
                                &(qmx_ch[ich].local_ep));
    rc_check(&rc,"mx_open_endpoint");
  }

  sleep(QMX_OPEN_WAITTIME);
  qmx_connect();

#ifdef _DEBUG_
  for (int ich = 0; ich < QMX_NUM_CH; ich++) {
    uint64_t  local_nic_id;
    int board = ich + 3;
    
    rc = mx_board_number_to_nic_id(board,&local_nic_id);
    rc_check(&rc,"mx_board_number_to_nic_id");
    printf("%s %12.12lX : ",qmx_hname,local_nic_id);
    printf("%d %12.12lX %d %12.12lX\n",qmx_ch[ich].local_rank, qmx_ch[ich].local_nic_id,
                                       qmx_ch[ich].remote_rank,qmx_ch[ich].remote_nic_id);
  }
#endif
}

extern "C"
void qmx_finalize() {
  mx_return_t rc;
  for (int ich = 0; ich < QMX_NUM_CH; ich++) {
    rc = mx_close_endpoint( qmx_ch[ich].local_ep );
    rc_check(&rc,"mx_close_endpoint");
  }
  rc = mx_finalize();
  rc_check(&rc,"mx_finalize");
}

static void 
qmx_ring_transmission(const int direction, const int byte, void *sbuff, void *rbuff) {

// direction = +1   0-(ch0)->2-(ch1)->1-(ch0)->3-(ch1)->0
// direction = -1   0-(ch1)->3-(ch0)->1-(ch1)->2-(ch0)->0

  mx_return_t rc;
  mx_segment_t recv_seg,send_seg;
  mx_request_t recv_req,send_req;
  mx_status_t status;
  uint32_t result;

  int parity = __qmx_get_parity(qmx_local_rank);

  int to_rank,from_rank;

  if (QMX_RING_ASCEND == direction) {

    to_rank   = __qmx_get_next_rank(qmx_local_rank);
    from_rank = __qmx_get_prev_rank(qmx_local_rank);

  } else if (QMX_RING_DESCEND == direction) {

    to_rank   = __qmx_get_prev_rank(qmx_local_rank);
    from_rank = __qmx_get_next_rank(qmx_local_rank);

  } else {

    fprintf(stderr,"qmx_ring_transmission : error in direction.\n");
    mx_finalize();
    exit(EXIT_FAILURE);

  }

  int to_ch   = qmx_get_channel_id(to_rank);
  int from_ch = qmx_get_channel_id(from_rank);

  if (to_ch == QMX_CHANNEL_ERROR) {
    mx_finalize();
    exit(EXIT_FAILURE);
  }
  if (from_ch == QMX_CHANNEL_ERROR) {
    mx_finalize();
    exit(EXIT_FAILURE);
  }

  send_seg.segment_ptr    = sbuff;
  send_seg.segment_length = byte;
  recv_seg.segment_ptr    = rbuff;
  recv_seg.segment_length = byte;

  if (0 == parity) {    

    rc = mx_isend(qmx_ch[to_ch].local_ep,&send_seg,1,qmx_ch[to_ch].remote_ep_addr,0x0L,NULL,&send_req);
    rc_check(&rc,"mx_isend");
    rc = mx_irecv(qmx_ch[from_ch].local_ep,&recv_seg,1,0x0L,0x0L,NULL,&recv_req);
    rc_check(&rc,"mx_irecv");

    rc =  mx_wait(qmx_ch[to_ch].local_ep,&send_req,QMX_TIMEOUT,&status,&result);
    rc_check(&rc,"mx_wait");
    rc =  mx_wait(qmx_ch[from_ch].local_ep,&recv_req,QMX_TIMEOUT,&status,&result);
    rc_check(&rc,"mx_wait");

  } else if (1 == parity) {

    rc = mx_irecv(qmx_ch[from_ch].local_ep,&recv_seg,1,0x0L,0x0L,NULL,&recv_req);
    rc_check(&rc,"mx_irecv");
    rc = mx_isend(qmx_ch[to_ch].local_ep,&send_seg,1,qmx_ch[to_ch].remote_ep_addr,0x0L,NULL,&send_req);
    rc_check(&rc,"mx_isend");

    rc =  mx_wait(qmx_ch[from_ch].local_ep,&recv_req,QMX_TIMEOUT,&status,&result);
    rc_check(&rc,"mx_wait");
    rc =  mx_wait(qmx_ch[to_ch].local_ep,&send_req,QMX_TIMEOUT,&status,&result);
    rc_check(&rc,"mx_wait");

  }


}

extern "C"
void qmx_barrier(){
  char sdat,rdat;
  sdat = (char)qmx_local_rank;

  rdat = -1;
  qmx_ring_transmission(QMX_RING_ASCEND,sizeof(char),&sdat,&rdat);
  qmx_ring_transmission(QMX_RING_ASCEND,sizeof(char),&rdat,&sdat);
  qmx_ring_transmission(QMX_RING_ASCEND,sizeof(char),&sdat,&rdat);
  qmx_ring_transmission(QMX_RING_ASCEND,sizeof(char),&rdat,&sdat);

#ifdef _DEBUG_
  if (sdat != qmx_local_rank) {
    fprintf(stderr,"qmx_barrier : error.\n");
    mx_finalize();
    exit(EXIT_FAILURE);
  }
#endif
}

template <typename T>
static void qmx_sendrecv(const int size, T sdata[], const int to_rank, T rdata[], const int from_rank){
  unsigned int parity = __qmx_get_parity(qmx_local_rank);
  int to_ch   = qmx_get_channel_id(to_rank);
  int from_ch = qmx_get_channel_id(from_rank);

  if (to_ch == QMX_CHANNEL_ERROR) {
    mx_finalize();
    exit(EXIT_FAILURE);
  }
  if (from_ch == QMX_CHANNEL_ERROR) {
    mx_finalize();
    exit(EXIT_FAILURE);
  }

  mx_return_t rc;
  mx_segment_t send_seg,recv_seg;
  mx_request_t send_req,recv_req;
  mx_status_t status;
  uint32_t result;

  send_seg.segment_ptr    = (void *)sdata;
  send_seg.segment_length = sizeof(T)*size;
  recv_seg.segment_ptr    = (void *)rdata;
  recv_seg.segment_length = sizeof(T)*size;

  if (0 == parity) {

    rc = mx_isend( qmx_ch[to_ch].local_ep, &send_seg, 1, qmx_ch[to_ch].remote_ep_addr, 0x0L, NULL, &send_req);
    rc_check(&rc,"mx_isend");
    rc = mx_irecv( qmx_ch[from_ch].local_ep, &recv_seg, 1, 0x0L, 0x0L, NULL, &recv_req);
    rc_check(&rc,"mx_irecv");

    rc =  mx_wait( qmx_ch[to_ch].local_ep, &send_req, QMX_TIMEOUT, &status, &result);
    rc_check(&rc,"mx_wait");
    rc =  mx_wait( qmx_ch[from_ch].local_ep, &recv_req, QMX_TIMEOUT, &status, &result);
    rc_check(&rc,"mx_wait");

  } else if (1 == parity) {

    rc = mx_irecv( qmx_ch[from_ch].local_ep, &recv_seg, 1, 0x0L, 0x0L, NULL, &recv_req);
    rc_check(&rc,"mx_irecv");
    rc = mx_isend( qmx_ch[to_ch].local_ep, &send_seg, 1, qmx_ch[to_ch].remote_ep_addr, 0x0L, NULL, &send_req);
    rc_check(&rc,"mx_isend");

    rc =  mx_wait( qmx_ch[from_ch].local_ep, &recv_req, QMX_TIMEOUT, &status, &result);
    rc_check(&rc,"mx_wait");
    rc =  mx_wait( qmx_ch[to_ch].local_ep, &send_req, QMX_TIMEOUT, &status, &result);
    rc_check(&rc,"mx_wait");

  }
}

template <typename T>
static void qmx_bcast(const int size, T data[], const int rank){
  T buff[QMX_NUM_RANK][size];
  int src_rank = __qmx_get_local_rank();
  int dst_rank;

  if ( src_rank == rank ) {
#pragma omp parallel for
    for (int i = 0; i < size; i++) {
      buff[src_rank][i] = data[i];
    }
  } else {
    for (int i = 0; i < size; i++) {
      buff[src_rank][i] = (T)(0);
    }
  }

  for (int i = 0; i < QMX_NUM_RANK-1; i++) {
    dst_rank = __qmx_get_1d_prev_rank(src_rank);
    qmx_ring_transmission(QMX_RING_ASCEND,size*sizeof(T),(void *)(buff[src_rank]),(void *)(buff[dst_rank]));
    src_rank = dst_rank;
  }

#pragma omp parallel for
  for (int i = 0; i < size; i++) {
    data[i] = buff[rank][i];
  }

}

template <typename T>
static void qmx_sumcast(const int size, T data[]){
  T buff[QMX_NUM_RANK][size];
  int src_rank = __qmx_get_local_rank();
  int dst_rank;

#pragma omp parallel for
  for (int i = 0; i < size; i++) {
    buff[src_rank][i] = data[i];
  }

  for (int i = 0; i < QMX_NUM_RANK-1; i++) {
    dst_rank = __qmx_get_1d_prev_rank(src_rank);
    qmx_ring_transmission(QMX_RING_ASCEND,size*sizeof(T),(void *)(buff[src_rank]),(void *)(buff[dst_rank]));
    src_rank = dst_rank;
  }

  for (int rank = 1; rank < QMX_NUM_RANK; rank++) {
#pragma omp parallel for
    for (int i = 0; i < size; i++) {
      buff[0][i] += buff[rank][i];
    }
  }
#pragma omp parallel for
  for (int i = 0; i < size; i++) {
    data[i] = buff[0][i];
  }
}

template <typename T>
static void qmx_allgather(const int size, T send[], T recv[]){
  T* buff_ptr;
  int src_rank = __qmx_get_local_rank();

#pragma omp parallel for
  for (int i = 0; i < size; i++) {
    recv[src_rank*size+i] = send[i];
  }

  for (int i = 0; i < QMX_NUM_RANK-1; i++) {
    buff_ptr = &recv[src_rank*size];
    src_rank = __qmx_get_1d_prev_rank(src_rank);
    qmx_ring_transmission(QMX_RING_ASCEND,size*sizeof(T),(void *)buff_ptr,(void *)&(recv[src_rank*size]));
  }

}

extern "C"
void qmx_bcast_int(int size, int *sdat, int rank) {
  qmx_bcast<int>(size, sdat, rank);
}

extern "C"
void qmx_bcast_char(int size, char *sdat, int rank) {
  qmx_bcast<char>(size, sdat, rank);
}

extern "C"
void qmx_bcast_float(int size, float *sdat, int rank) {
  qmx_bcast<float>(size, sdat, rank);
}

extern "C"
void qmx_bcast_double(int size, double *sdat, int rank) {
  qmx_bcast<double>(size, sdat, rank);
}

extern "C"
void qmx_bcast_cmplx(int size, std::complex<float> *sdat, int rank) {
  qmx_bcast< std::complex<float> >(size, sdat, rank);
}

extern "C"
void qmx_bcast_dcmplx(int size, std::complex<double> *sdat, int rank) {
  qmx_bcast< std::complex<double> >(size, sdat, rank);
}

extern "C"
void qmx_sumcast_int(int size, int *sdat) {
  qmx_sumcast<int>(size, sdat);
}

extern "C"
void qmx_sumcast_char(int size, char *sdat) {
  qmx_sumcast<char>(size, sdat);
}

extern "C"
void qmx_sumcast_float(int size, float *sdat) {
  qmx_sumcast<float>(size, sdat);
}

extern "C"
void qmx_sumcast_double(int size, double *sdat) {
  qmx_sumcast<double>(size, sdat);
}

extern "C"
void qmx_sumcast_cmplx(int size, std::complex<float> *sdat) {
  qmx_sumcast< std::complex<float> >(size, sdat);
}

extern "C"
void qmx_sumcast_dcmplx(int size, std::complex<double> *sdat) {
  qmx_sumcast< std::complex<double> >(size, sdat);
}

extern "C"
void qmx_allgather_int(const int size, int *send, int *recv){
  qmx_allgather<int>(size,send,recv);
}

extern "C"
void qmx_allgather_char(const int size, char *send, char *recv){
  qmx_allgather<char>(size,send,recv);
}

extern "C"
void qmx_allgather_float(const int size, float *send, float *recv){
  qmx_allgather<float>(size,send,recv);
}

extern "C"
void qmx_allgather_double(const int size, double *send, double *recv){
  qmx_allgather<double>(size,send,recv);
}

extern "C"
void qmx_allgather_cmplx(const int size, std::complex<float> *send, std::complex<float> *recv){
  qmx_allgather< std::complex<float> >(size,send,recv);
}

extern "C"
void qmx_allgather_dcmplx(const int size, std::complex<double> *send, std::complex<double> *recv){
  qmx_allgather< std::complex<double> >(size,send,recv);
}

extern "C"
void qmx_sendrecv_int(int size, int *sdat, int to_rank, int *rdat, int from_rank) {
  qmx_sendrecv<int>(size, sdat, to_rank, rdat, from_rank);
}

extern "C"
void qmx_sendrecv_char(int size, char *sdat, int to_rank, char *rdat, int from_rank) {
  qmx_sendrecv<char>(size, sdat, to_rank, rdat, from_rank);
}

extern "C"
void qmx_sendrecv_float(int size, float *sdat, int to_rank, float *rdat, int from_rank) {
  qmx_sendrecv<float>(size, sdat, to_rank, rdat, from_rank);
}

extern "C"
void qmx_sendrecv_double(int size, double *sdat, int to_rank, double *rdat, int from_rank) {
  qmx_sendrecv<double>(size, sdat, to_rank, rdat, from_rank);
}

extern "C"
void qmx_sendrecv_cmplx(int size, std::complex<float> *sdat, int to_rank, std::complex<float> *rdat, int from_rank) {
  qmx_sendrecv< std::complex<float> >(size, sdat, to_rank, rdat, from_rank);
}

extern "C"
void qmx_sendrecv_dcmplx(int size, std::complex<double> *sdat, int to_rank, std::complex<double> *rdat, int from_rank) {
  qmx_sendrecv< std::complex<double> >(size, sdat, to_rank, rdat, from_rank);
}
